@@ -27,6 +27,7 @@
 #include <QCryptographicHash>
 #include <QFile>
 #include <QFileInfo>
+#include <QScopeGuard>
 #include <QStandardPaths>
 #include <QStringEncoder>
 #include <QTemporaryFile>
@@ -76,34 +77,60 @@ TextBuffer::~TextBuffer()
     // not allowed during editing
     Q_ASSERT(m_editingTransactions == 0);
 
-    // kill all ranges, work on copy, they will remove themself from the hash
-    QSet<TextRange *> copyRanges = m_ranges;
-    qDeleteAll(copyRanges);
-    Q_ASSERT(m_ranges.empty());
+    // invalidate all moving stuff
+    std::vector<Kate::TextRange *> rangesWithFeedback;
+    for (auto b : m_blocks) {
+        auto cursors = std::move(b->m_cursors);
+        for (auto it = cursors.begin(); it != cursors.end(); ++it) {
+            auto cursor = *it;
+            // update the block
+            cursor->m_block = nullptr;
+            cursor->m_line = cursor->m_column = -1;
+            cursor->m_buffer = nullptr;
+            if (auto r = cursor->kateRange()) {
+                r->m_buffer = nullptr;
+                if (r->feedback()) {
+                    rangesWithFeedback.push_back(r);
+                }
+            }
+        }
+    }
+
+    // uniquify ranges
+    std::sort(rangesWithFeedback.begin(), rangesWithFeedback.end());
+    auto it = std::unique(rangesWithFeedback.begin(), rangesWithFeedback.end());
+    std::for_each(rangesWithFeedback.begin(), it, [](Kate::TextRange *range) {
+        range->feedback()->rangeInvalid(range);
+    });
 
     // clean out all cursors and lines, only cursors belonging to range will survive
-    for (TextBlock *block : std::as_const(m_blocks)) {
-        block->deleteBlockContent();
+    for (TextBlock *block : m_blocks) {
+        block->clearLines();
     }
 
     // delete all blocks, now that all cursors are really deleted
     // else asserts in destructor of blocks will fail!
     qDeleteAll(m_blocks);
     m_blocks.clear();
-
-    // kill all invalid cursors, do this after block deletion, to uncover if they might be still linked in blocks
-    QSet<TextCursor *> copyCursors = m_invalidCursors;
-    qDeleteAll(copyCursors);
-    Q_ASSERT(m_invalidCursors.empty());
 }
 
 void TextBuffer::invalidateRanges()
 {
-    // invalidate all ranges, work on copy, they might delete themself...
-    const QSet<TextRange *> copyRanges = m_ranges;
-    for (TextRange *range : copyRanges) {
-        range->setRange(KTextEditor::Cursor::invalid(), KTextEditor::Cursor::invalid());
+    std::vector<Kate::TextRange *> ranges;
+    ranges.reserve(m_blocks.size());
+    for (TextBlock *block : m_blocks) {
+        for (auto cursor : block->m_cursors) {
+            if (cursor->kateRange()) {
+                ranges.push_back(cursor->kateRange());
+            }
+        }
     }
+    // uniquify ranges
+    std::sort(ranges.begin(), ranges.end());
+    auto it = std::unique(ranges.begin(), ranges.end());
+    std::for_each(ranges.begin(), it, [](Kate::TextRange *range) {
+        range->setRange({KTextEditor::Cursor::invalid(), KTextEditor::Cursor::invalid()});
+    });
 }
 
 void TextBuffer::clear()
@@ -111,25 +138,38 @@ void TextBuffer::clear()
     // not allowed during editing
     Q_ASSERT(m_editingTransactions == 0);
 
+    m_multilineRanges.clear();
     invalidateRanges();
 
     // new block for empty buffer
     TextBlock *newBlock = new TextBlock(this, 0);
     newBlock->appendLine(QString());
 
-    // clean out all cursors and lines, either move them to newBlock or invalidate them, if belonging to a range
+    // clean out all cursors and lines, move them to newBlock if not belonging to a range
     for (TextBlock *block : std::as_const(m_blocks)) {
-        block->clearBlockContent(newBlock);
+        auto cursors = std::move(block->m_cursors);
+        for (auto it = cursors.begin(); it != cursors.end(); ++it) {
+            auto cursor = *it;
+            if (!cursor->kateRange()) {
+                // update the block
+                cursor->m_block = newBlock;
+                // move the cursor into the target block
+                cursor->m_line = cursor->m_column = 0;
+                newBlock->m_cursors.push_back(cursor);
+                // remove it and advance to next element
+            }
+            // skip cursors with ranges, we need to invalidate the ranges later
+        }
+        block->clearLines();
     }
+    std::sort(newBlock->m_cursors.begin(), newBlock->m_cursors.end());
 
     // kill all buffer blocks
     qDeleteAll(m_blocks);
-    m_blocks.clear();
-    m_startLines.clear();
-
     // insert one block with one empty line
-    m_blocks.push_back(newBlock);
-    m_startLines.push_back(0);
+    m_blocks = {newBlock};
+    m_startLines = {0};
+    m_blockSizes = {1};
 
     // reset lines and last used block
     m_lines = 1;
@@ -176,8 +216,8 @@ int TextBuffer::cursorToOffset(KTextEditor::Cursor c) const
 
     int off = 0;
     const int blockIndex = blockForLine(c.line());
-    for (auto it = m_blocks.begin(), end = m_blocks.begin() + blockIndex; it != end; ++it) {
-        off += (*it)->blockSize();
+    for (auto it = m_blockSizes.begin(), end = m_blockSizes.begin() + blockIndex; it != end; ++it) {
+        off += *it;
     }
 
     auto block = m_blocks[blockIndex];
@@ -199,10 +239,12 @@ KTextEditor::Cursor TextBuffer::offsetToCursor(int offset) const
 {
     if (offset >= 0) {
         int off = 0;
-        for (auto block : m_blocks) {
-            if (off + block->blockSize() < offset) {
-                off += block->blockSize();
+        int blockIdx = 0;
+        for (int blockSize : m_blockSizes) {
+            if (off + blockSize < offset) {
+                off += blockSize;
             } else {
+                auto block = m_blocks[blockIdx];
                 const int lines = block->lines();
                 int start = block->startLine();
                 int end = start + lines;
@@ -214,6 +256,7 @@ KTextEditor::Cursor TextBuffer::offsetToCursor(int offset) const
                     off += len + 1;
                 }
             }
+            blockIdx++;
         }
     }
     return KTextEditor::Cursor::invalid();
@@ -223,8 +266,8 @@ QString TextBuffer::text() const
 {
     QString text;
     qsizetype size = 0;
-    for (auto b : m_blocks) {
-        size += b->blockSize();
+    for (int blockSize : m_blockSizes) {
+        size += blockSize;
     }
     text.reserve(size);
     size -= 1; // remove -1, last newline
@@ -305,6 +348,7 @@ void TextBuffer::wrapLine(const KTextEditor::Cursor position)
     // this call will trigger fixStartLines
     ++m_lines; // first alter the line counter, as functions called will need the valid one
     m_blocks.at(blockIndex)->wrapLine(position, blockIndex);
+    m_blockSizes[blockIndex] += 1;
 
     // remember changes
     ++m_revision;
@@ -398,6 +442,7 @@ void TextBuffer::insertText(const KTextEditor::Cursor position, const QString &t
 
     // let the block handle the insertText
     m_blocks.at(blockIndex)->insertText(position, text);
+    m_blockSizes[blockIndex] += text.size();
 
     // remember changes
     ++m_revision;
@@ -441,6 +486,7 @@ void TextBuffer::removeText(KTextEditor::Range range)
     // let the block handle the removeText, retrieve removed text
     QString text;
     m_blocks.at(blockIndex)->removeText(range, text);
+    m_blockSizes[blockIndex] -= text.size();
 
     // remember changes
     ++m_revision;
@@ -508,6 +554,12 @@ void TextBuffer::fixStartLines(int startBlock, int value)
 
 void TextBuffer::balanceBlock(int index)
 {
+    auto check = qScopeGuard([this] {
+        if (!(m_blocks.size() == m_startLines.size() && m_blocks.size() == m_blockSizes.size())) {
+            qFatal("blocks/startlines/blocksizes are not equal in size!");
+        }
+    });
+
     // two cases, too big or too small block
     TextBlock *blockToBalance = m_blocks.at(index);
 
@@ -521,6 +573,7 @@ void TextBuffer::balanceBlock(int index)
         TextBlock *newBlock = new TextBlock(this, index + 1);
         m_blocks.insert(m_blocks.begin() + index + 1, newBlock);
         m_startLines.insert(m_startLines.begin() + index + 1, newBlockStartLine);
+        m_blockSizes.insert(m_blockSizes.begin() + index + 1, 0);
 
         // adjust block indexes
         for (auto it = m_blocks.begin() + index, end = m_blocks.end(); it != end; ++it) {
@@ -542,6 +595,7 @@ void TextBuffer::balanceBlock(int index)
         if (blockToBalance->lines() == 0) {
             m_blocks.erase(m_blocks.begin());
             m_startLines.erase(m_startLines.begin());
+            m_blockSizes.erase(m_blockSizes.begin());
             Q_ASSERT(m_startLines[0] == 0);
             for (auto it = m_blocks.begin(), end = m_blocks.end(); it != end; ++it) {
                 (*it)->setBlockIndex(index++);
@@ -560,18 +614,19 @@ void TextBuffer::balanceBlock(int index)
 
     // merge block
     blockToBalance->mergeBlock(targetBlock);
+    m_blockSizes[index - 1] += m_blockSizes[index];
 
     // delete old block
     delete blockToBalance;
     m_blocks.erase(m_blocks.begin() + index);
     m_startLines.erase(m_startLines.begin() + index);
+    m_blockSizes.erase(m_blockSizes.begin() + index);
 
     for (auto it = m_blocks.begin() + index, end = m_blocks.end(); it != end; ++it) {
         (*it)->setBlockIndex(index++);
     }
 
     Q_ASSERT(index == (int)m_blocks.size());
-    Q_ASSERT(m_blocks.size() == m_startLines.size());
 }
 
 void TextBuffer::debugPrint(const QString &title) const
@@ -613,10 +668,12 @@ bool TextBuffer::load(const QString &filename, bool &encodingErrors, bool &tooLo
         }
         m_blocks.resize(1);
         m_startLines.resize(1);
+        m_blockSizes.resize(1);
 
         // remove lines in first block
         m_blocks.back()->clearLines();
         m_startLines.back() = 0;
+        m_blockSizes.back() = 0;
         m_lines = 0;
 
         // reset error flags
@@ -638,6 +695,7 @@ bool TextBuffer::load(const QString &filename, bool &encodingErrors, bool &tooLo
             // create one dummy textline, in any case
             m_blocks.back()->appendLine(QString());
             m_lines++;
+            m_blockSizes[0] = 1;
             return false;
         }
 
@@ -662,10 +720,12 @@ bool TextBuffer::load(const QString &filename, bool &encodingErrors, bool &tooLo
                 int startLine = m_blocks.back()->startLine() + m_blocks.back()->lines();
                 m_blocks.push_back(new TextBlock(this, index));
                 m_startLines.push_back(startLine);
+                m_blockSizes.push_back(0);
             }
 
             // append line to last block
             m_blocks.back()->appendLine(QString(file.unicode() + offset, length));
+            m_blockSizes.back() += length + 1;
             ++m_lines;
         }
 
@@ -747,14 +807,20 @@ bool TextBuffer::save(const QString &filename)
     // codec must be set, else below we fail!
     Q_ASSERT(!m_textCodec.isEmpty());
 
-    SaveResult saveRes = saveBufferUnprivileged(filename);
+    // ensure we do not kill symlinks, see bug 498589
+    auto realFile = filename;
+    if (const auto realFileResolved = QFileInfo(realFile).canonicalFilePath(); !realFileResolved.isEmpty()) {
+        realFile = realFileResolved;
+    }
 
+    const auto saveRes = saveBufferUnprivileged(realFile);
     if (saveRes == SaveResult::Failed) {
         return false;
-    } else if (saveRes == SaveResult::MissingPermissions) {
+    }
+    if (saveRes == SaveResult::MissingPermissions) {
         // either unit-test mode or we're missing permissions to write to the
         // file => use temporary file and try to use authhelper
-        if (!saveBufferEscalated(filename)) {
+        if (!saveBufferEscalated(realFile)) {
             return false;
         }
     }
@@ -935,7 +1001,7 @@ bool TextBuffer::saveBufferEscalated(const QString &filename)
 #endif
 }
 
-void TextBuffer::notifyAboutRangeChange(KTextEditor::View *view, KTextEditor::LineRange lineRange, bool needsRepaint)
+void TextBuffer::notifyAboutRangeChange(KTextEditor::View *view, KTextEditor::LineRange lineRange, bool needsRepaint, TextRange *deleteRange)
 {
     // ignore calls if no document is around
     if (!m_document) {
@@ -947,12 +1013,12 @@ void TextBuffer::notifyAboutRangeChange(KTextEditor::View *view, KTextEditor::Li
     const QList<KTextEditor::View *> views = m_document->views();
     for (KTextEditor::View *curView : views) {
         // filter wrong views
-        if (view && view != curView) {
+        if (view && view != curView && !deleteRange) {
             continue;
         }
 
         // notify view, it is really a kate view
-        static_cast<KTextEditor::ViewPrivate *>(curView)->notifyAboutRangeChange(lineRange, needsRepaint);
+        static_cast<KTextEditor::ViewPrivate *>(curView)->notifyAboutRangeChange(lineRange, needsRepaint, deleteRange);
     }
 }
 
@@ -962,6 +1028,56 @@ void TextBuffer::markModifiedLinesAsSaved()
         block->markModifiedLinesAsSaved();
     }
 }
+
+void TextBuffer::addMultilineRange(TextRange *range)
+{
+    auto it = std::find(m_multilineRanges.begin(), m_multilineRanges.end(), range);
+    if (it == m_multilineRanges.end()) {
+        m_multilineRanges.push_back(range);
+        return;
+    }
+}
+
+void TextBuffer::removeMultilineRange(TextRange *range)
+{
+    m_multilineRanges.erase(std::remove(m_multilineRanges.begin(), m_multilineRanges.end(), range), m_multilineRanges.end());
+}
+
+bool TextBuffer::hasMultlineRange(KTextEditor::MovingRange *range) const
+{
+    return std::find(m_multilineRanges.begin(), m_multilineRanges.end(), range) != m_multilineRanges.end();
+}
+
+void TextBuffer::rangesForLine(int line, KTextEditor::View *view, bool rangesWithAttributeOnly, QList<TextRange *> &outRanges) const
+{
+    outRanges.clear();
+    // get block, this will assert on invalid line
+    const int blockIndex = blockForLine(line);
+    m_blocks.at(blockIndex)->rangesForLine(line, view, rangesWithAttributeOnly, outRanges);
+    // printf("Requested range for line %d, available %d\n", (int)line, (int)m_multilineRanges.size());
+    for (TextRange *range : std::as_const(m_multilineRanges)) {
+        if (rangesWithAttributeOnly && !range->hasAttribute()) {
+            continue;
+        }
+
+        // we want ranges for no view, but this one's attribute is only valid for views
+        if (!view && range->attributeOnlyForViews()) {
+            continue;
+        }
+
+        // the range's attribute is not valid for this view
+        if (range->view() && range->view() != view) {
+            continue;
+        }
+
+        // if line is in the range, ok
+        if (range->startInternal().lineInternal() <= line && line <= range->endInternal().lineInternal()) {
+            outRanges.append(range);
+        }
+    }
+    std::sort(outRanges.begin(), outRanges.end());
+    outRanges.erase(std::unique(outRanges.begin(), outRanges.end()), outRanges.end());
+}
 }
 
 #include "moc_katetextbuffer.cpp"
