@@ -715,7 +715,7 @@ void KateRenderer::paintTextLine(QPainter &paint,
         paint.drawLine(0, (lineHeight() * range->viewLineCount()) - 2, xEnd - xStart, (lineHeight() * range->viewLineCount()) - 2);
     }
 
-    if (range->layout()) {
+    if (range->layout().lineCount() > 0) {
         bool drawSelection =
             m_view && m_view->selection() && showSelections() && m_view->selectionRange().overlapsLine(range->line()) && !flags.testFlag(SkipDrawLineSelection);
         // Draw selection in block selection mode. We need 2 kinds of selections that QTextLayout::draw can't render:
@@ -749,7 +749,7 @@ void KateRenderer::paintTextLine(QPainter &paint,
             paint.setPen(attribute(KSyntaxHighlighting::Theme::TextStyle::Normal)->foreground().color());
             // Draw the text :)
 
-            if (range->layout()->textOption().textDirection() == Qt::RightToLeft) {
+            if (range->layout().textOption().textDirection() == Qt::RightToLeft) {
                 // If the text is RTL, we draw text background ourselves
                 auto decos = decorationsForLine(range->textLine(), range->line(), false);
                 auto sr = view()->selectionRange();
@@ -772,10 +772,10 @@ void KateRenderer::paintTextLine(QPainter &paint,
                     paintTextBackground(paint, range, additionalFormats, config()->selectionColor(), xStart);
                 }
                 // DONT apply clipping, it breaks rendering when there are selections
-                range->layout()->draw(&paint, QPoint(-xStart, 0), additionalFormats);
+                range->layout().draw(&paint, QPoint(-xStart, 0), additionalFormats);
 
             } else {
-                range->layout()->draw(&paint, QPoint(-xStart, 0), QList<QTextLayout::FormatRange>{}, textClipRect);
+                range->layout().draw(&paint, QPoint(-xStart, 0), QList<QTextLayout::FormatRange>{}, textClipRect);
             }
         }
 
@@ -841,7 +841,7 @@ void KateRenderer::paintTextLine(QPainter &paint,
                 if (spaceIndex >= trailingPos) {
                     QVarLengthArray<int, 32> spacePositions;
                     // Adjust to visible contents
-                    const auto dir = range->layout()->textOption().textDirection();
+                    const auto dir = range->layout().textOption().textDirection();
                     const bool isRTL = dir == Qt::RightToLeft && m_view->dynWordWrap();
                     int start = isRTL ? xEnd : xStart;
                     int end = isRTL ? xStart : xEnd;
@@ -874,7 +874,7 @@ void KateRenderer::paintTextLine(QPainter &paint,
                         const int spaceIdx = *rit;
                         qreal x = line.lineLayout().cursorToX(spaceIdx) - xStart;
                         int dir = 1; // 1 == ltr, -1 == rtl
-                        if (range->layout()->textOption().alignment() == Qt::AlignRight) {
+                        if (range->layout().textOption().alignment() == Qt::AlignRight) {
                             dir = -1;
                             if (spaceIdx > 0) {
                                 QChar c = text.at(spaceIdx - 1);
@@ -912,7 +912,7 @@ void KateRenderer::paintTextLine(QPainter &paint,
                 const int y = lineHeight() * i + m_fontAscent;
 
                 static const QRegularExpression nonPrintableSpacesRegExp(
-                    QStringLiteral("[\\x{2000}-\\x{200F}\\x{2028}-\\x{202F}\\x{205F}-\\x{2064}\\x{206A}-\\x{206F}]"));
+                    QStringLiteral("[\\x{0000}-\\x{0008}\\x{000A}-\\x{001F}\\x{2000}-\\x{200F}\\x{2028}-\\x{202F}\\x{205F}-\\x{2064}\\x{206A}-\\x{206F}]"));
                 QRegularExpressionMatchIterator i = nonPrintableSpacesRegExp.globalMatch(text, line.lineLayout().xToCursor(xStart));
 
                 while (i.hasNext()) {
@@ -947,9 +947,17 @@ void KateRenderer::paintTextLine(QPainter &paint,
             auto mIt = std::lower_bound(secCursors.begin(), secCursors.end(), range->line(), [](const KTextEditor::ViewPrivate::SecondaryCursor &l, int line) {
                 return l.pos->line() < line;
             });
+            bool skipPrimary = false;
             if (mIt != secCursors.end() && mIt->cursor().line() == range->line()) {
+                KTextEditor::Cursor last = KTextEditor::Cursor::invalid();
+                auto primaryCursor = *cursor;
                 for (; mIt != secCursors.end(); ++mIt) {
                     auto cursor = mIt->cursor();
+                    skipPrimary = skipPrimary || cursor == primaryCursor;
+                    if (cursor == last) {
+                        continue;
+                    }
+                    last = cursor;
                     if (cursor.line() == range->line()) {
                         paintCaret(cursor, range, paint, xStart, xEnd);
                     } else {
@@ -957,7 +965,9 @@ void KateRenderer::paintTextLine(QPainter &paint,
                     }
                 }
             }
-            paintCaret(*cursor, range, paint, xStart, xEnd);
+            if (!skipPrimary) {
+                paintCaret(*cursor, range, paint, xStart, xEnd);
+            }
         }
     }
 
@@ -979,7 +989,7 @@ void KateRenderer::paintTextLine(QPainter &paint,
             const int viewLine = range->viewLineForColumn(column);
             // We only consider a line "rtl" if dynamic wrap is enabled. If it is disabled, our
             // text is always on the left side of the view
-            const auto dir = range->layout()->textOption().textDirection();
+            const auto dir = range->layout().textOption().textDirection();
             const bool isRTL = dir == Qt::RightToLeft && m_view->dynWordWrap();
 
             // Determine the position where to paint the note.
@@ -1043,7 +1053,7 @@ void KateRenderer::paintCaret(KTextEditor::Cursor cursor, KateLineLayout *range,
         int caretWidth;
         int lineWidth = 2;
         QColor color;
-        QTextLine line = range->layout()->lineForTextPosition(qMin(cursor.column(), range->length()));
+        QTextLine line = range->layout().lineForTextPosition(qMin(cursor.column(), range->length()));
 
         // Determine the caret's style
         KTextEditor::caretStyles style = caretStyle();
@@ -1067,7 +1077,7 @@ void KateRenderer::paintCaret(KTextEditor::Cursor cursor, KateLineLayout *range,
             color = m_caretOverrideColor;
         } else {
             // search for the FormatRange that includes the cursor
-            const auto formatRanges = range->layout()->formats();
+            const auto formatRanges = range->layout().formats();
             for (const QTextLayout::FormatRange &r : formatRanges) {
                 if ((r.start <= cursor.column()) && ((r.start + r.length) > cursor.column())) {
                     // check for Qt::NoBrush, as the returned color is black() and no invalid QColor
@@ -1104,7 +1114,7 @@ void KateRenderer::paintCaret(KTextEditor::Cursor cursor, KateLineLayout *range,
 
         if (cursor.column() <= range->length()) {
             // Ensure correct cursor placement for RTL text
-            if (range->layout()->textOption().textDirection() == Qt::RightToLeft) {
+            if (range->layout().textOption().textDirection() == Qt::RightToLeft) {
                 xStart += caretWidth;
             }
             qreal width = 0;
@@ -1117,7 +1127,7 @@ void KateRenderer::paintCaret(KTextEditor::Cursor cursor, KateLineLayout *range,
                     }
                 }
             }
-            drawCursor(*range->layout(), &paint, QPoint(-xStart - width, 0), cursor.column(), caretWidth, lineHeight());
+            drawCursor(range->layout(), &paint, QPoint(-xStart - width, 0), cursor.column(), caretWidth, lineHeight());
         } else {
             // Off the end of the line... must be block mode. Draw the caret ourselves.
             const KateTextLayout &lastLine = range->viewLine(range->viewLineCount() - 1);
@@ -1252,15 +1262,10 @@ void KateRenderer::layoutLine(KateLineLayout *lineLayout, int maxwidth, bool cac
 
     Kate::TextLine textLine = lineLayout->textLine();
 
-    QTextLayout *l = lineLayout->layout();
-    if (!l) {
-        l = new QTextLayout(textLine.text(), m_font);
-    } else {
-        l->setText(textLine.text());
-        l->setFont(m_font);
-    }
-
-    l->setCacheEnabled(cacheLayout);
+    QTextLayout &l = lineLayout->modifiableLayout();
+    l.setText(textLine.text());
+    l.setFont(m_font);
+    l.setCacheEnabled(cacheLayout);
 
     // Initial setup of the QTextLayout.
 
@@ -1300,7 +1305,7 @@ void KateRenderer::layoutLine(KateLineLayout *lineLayout, int maxwidth, bool cac
         opt.setTextDirection(Qt::LeftToRight);
     }
 
-    l->setTextOption(opt);
+    l.setTextOption(opt);
 
     // Syntax highlighting, inbuilt and arbitrary
     QList<QTextLayout::FormatRange> decorations = decorationsForLine(textLine, lineLayout->line());
@@ -1344,19 +1349,19 @@ void KateRenderer::layoutLine(KateLineLayout *lineLayout, int maxwidth, bool cac
             // If it is outside of the text, we don't have to make space for it.
             if (column == 0) {
                 firstLineOffset = width;
-            } else if (column < l->text().length()) {
+            } else if (column < l.text().length()) {
                 QTextCharFormat text_char_format;
                 const qreal caretWidth = caretStyle() == KTextEditor::caretStyles::Line ? 2.0 : 0.0;
                 text_char_format.setFontLetterSpacing(width + caretWidth);
                 text_char_format.setFontLetterSpacingType(QFont::AbsoluteSpacing);
-                decorations.append(QTextLayout::FormatRange{.start=column - 1, .length=1, .format=text_char_format});
+                decorations.append(QTextLayout::FormatRange{.start = column - 1, .length = 1, .format = text_char_format});
             }
         }
     }
-    l->setFormats(decorations);
+    l.setFormats(decorations);
 
     // Begin layouting
-    l->beginLayout();
+    l.beginLayout();
 
     int height = 0;
     int shiftX = 0;
@@ -1364,7 +1369,7 @@ void KateRenderer::layoutLine(KateLineLayout *lineLayout, int maxwidth, bool cac
     bool needShiftX = (maxwidth != -1) && m_view && (m_view->config()->dynWordWrapAlignIndent() > 0);
 
     while (true) {
-        QTextLine line = l->createLine();
+        QTextLine line = l.createLine();
         if (!line.isValid()) {
             break;
         }
@@ -1403,9 +1408,8 @@ void KateRenderer::layoutLine(KateLineLayout *lineLayout, int maxwidth, bool cac
         height += lineHeight();
     }
 
-    l->endLayout();
-
-    lineLayout->setLayout(l);
+    // will end layout and trigger that we mark the layout as changed
+    lineLayout->endLayout();
 }
 
 // 1) QString::isRightToLeft() sux
@@ -1491,7 +1495,7 @@ void KateRenderer::setCaretOverrideColor(const QColor &color)
     m_caretOverrideColor = color;
 }
 
-void KateRenderer::paintSelection(QPaintDevice *d, int startLine, int xStart, int endLine, int xEnd, qreal scale)
+void KateRenderer::paintSelection(QPaintDevice *d, int startLine, int xStart, int endLine, int xEnd, int viewWidth, qreal scale)
 {
     if (!d || scale < 0.0) {
         return;
@@ -1523,13 +1527,13 @@ void KateRenderer::paintSelection(QPaintDevice *d, int startLine, int xStart, in
         // compute layout WITHOUT cache to not poison it + render it
         KateLineLayout lineLayout(*this);
         lineLayout.setLine(line, -1);
-        layoutLine(&lineLayout, -1 /* no wrap */, false /* no layout cache */);
+        layoutLine(&lineLayout, viewWidth, false /* no layout cache */);
         KateRenderer::PaintTextLineFlags flags;
         flags.setFlag(KateRenderer::SkipDrawFirstInvisibleLineUnderlined);
         flags.setFlag(KateRenderer::SkipDrawLineSelection);
         paintTextLine(paint, &lineLayout, 0, 0, QRectF{}, nullptr, flags);
 
         // translate for next line
-        paint.translate(0, lineHeight);
+        paint.translate(0, lineHeight * lineLayout.viewLineCount());
     }
 }
