@@ -330,13 +330,14 @@ KateDocumentConfig::KateDocumentConfig()
     addConfigEntry(ConfigEntry(SmartHome, "Smart Home", QString(), true));
     addConfigEntry(ConfigEntry(ShowTabs, "Show Tabs", QString(), true));
     addConfigEntry(ConfigEntry(IndentOnTab, "Indent On Tab", QString(), true));
-    addConfigEntry(ConfigEntry(KeepExtraSpaces, "Keep Extra Spaces", QString(), false));
+    addConfigEntry(ConfigEntry(KeepExtraSpaces, "Keep Extra Spaces", QStringLiteral("keep-extra-spaces"), false));
     addConfigEntry(ConfigEntry(BackspaceIndents, "Indent On Backspace", QString(), true));
     addConfigEntry(ConfigEntry(ShowSpacesMode, "Show Spaces", QString(), KateDocumentConfig::None));
     addConfigEntry(ConfigEntry(TrailingMarkerSize, "Trailing Marker Size", QString(), 1));
-    addConfigEntry(ConfigEntry(RemoveSpacesMode, "Remove Spaces", QString(), 1 /* on modified lines per default */, [](const QVariant &value) {
-        return inBounds(0, value, 2);
-    }));
+    addConfigEntry(
+        ConfigEntry(RemoveSpacesMode, "Remove Spaces", QStringLiteral("remove-spaces"), 1 /* on modified lines per default */, [](const QVariant &value) {
+            return inBounds(0, value, 2);
+        }));
     addConfigEntry(ConfigEntry(NewlineAtEOF, "Newline at End of File", QString(), true));
     addConfigEntry(ConfigEntry(OverwriteMode, "Overwrite Mode", QString(), false));
     addConfigEntry(
@@ -360,6 +361,8 @@ KateDocumentConfig::KateDocumentConfig()
 
     // Shall we do auto reloading for stuff e.g. in Git?
     addConfigEntry(ConfigEntry(AutoReloadIfStateIsInVersionControl, "Auto Reload If State Is In Version Control", QString(), true));
+    // .editorconfig
+    addConfigEntry(ConfigEntry(UseEditorConfig, "Use Editor Config", QString(), true));
 
     // finalize the entries, e.g. hashs them
     finalizeConfigEntries();
@@ -564,6 +567,7 @@ KateViewConfig::KateViewConfig()
     addConfigEntry(ConfigEntry(StatusbarLineColumnCompact, "Statusbar Line Column Compact Mode", QString(), true));
     addConfigEntry(ConfigEntry(ShowStatusbarEOL, "Shoe Line Ending Type in Statusbar", QString(), false));
     addConfigEntry(ConfigEntry(EnableAccessibility, "Enable Accessibility", QString(), true));
+    addConfigEntry(ConfigEntry(CycleThroughBookmarks, "Cycle Through Bookmarks", QString(), true));
 
     // Never forget to finalize or the <CommandName> becomes not available
     finalizeConfigEntries();
@@ -973,11 +977,6 @@ void KateRendererConfig::setFont(const QFont &font)
     configStart();
     m_font = font;
     m_fontSet = true;
-
-    // Set full hinting instead to ensure the letters are aligned properly, bug 482659
-    // https://codereview.qt-project.org/c/qt/qtbase/+/546168
-    m_font.setHintingPreference(QFont::PreferFullHinting);
-
     configEnd();
 }
 
